import numpy as np

from dune.perftool.generation import post_include

from dune.perftool.generation.loopy import DuneGlobalArg
from dune.perftool.loopy.temporary import DuneTemporaryVariable
from dune.perftool.loopy.vcl import VCLTypeRegistry
from dune.perftool.generation import (include_file,
                                      retrieve_cache_functions,
                                      )
from dune.perftool.options import get_option

from loopy.symbolic import Literal
from loopy.target import (TargetBase,
                          ASTBuilderBase,
                          DummyHostASTBuilder,
                          )
from loopy.target.c import CASTBuilder
from loopy.target.c.codegen.expression import ExpressionToCExpressionMapper, CExpressionToCodeMapper
from loopy.tools import is_integer
from loopy.types import NumpyType

from pymbolic.mapper.stringifier import PREC_NONE
import pymbolic.primitives as prim

import pytools as pt

import cgen


def _type_to_op_counter_type(name):
    return "oc::OpCounter<{}>".format(name)


@pt.memoize
def numpy_to_cpp_dtype(key):
    _registry = {'float32': 'float',
                 'int32': 'int',
                 'float64': 'double',
                 'string': 'std::string',
                 'str': 'std::string'}

    if get_option('opcounter'):
        _registry['float32'] = _type_to_op_counter_type('float')
        _registry['float64'] = _type_to_op_counter_type('double')

    return _registry[key]


class DuneExpressionToCExpressionMapper(ExpressionToCExpressionMapper):
    def map_subscript(self, expr, type_context):
        arr = self.find_array(expr)
        if isinstance(arr, (DuneTemporaryVariable, DuneGlobalArg)) and not arr.managed:
            # If there is but one index, we do not need to handle this
            if isinstance(expr.index, (prim.Variable, int)):
                return expr

            # Else, we construct a nested Subscript chain
            ret = expr.aggregate
            for i in expr.index:
                ret = prim.Subscript(ret, i)
            return ret
        else:
            return ExpressionToCExpressionMapper.map_subscript(self, expr, type_context)

    def map_floor_div(self, expr, enclosing_prec):
        # Loopy generates floor divs via a macro int_floor_div_pos_b
        # I have no idea why it does this and would like to generate
        # operator/ instead. At some point, I should ask Andreas about
        # the intention of that macro. Returning expr will pass this into
        # the CExpressionToCodeMapper as is => operator/.
        return expr

    def map_constant(self, expr, type_context):
        ret = ExpressionToCExpressionMapper.map_constant(self, expr, type_context)
        if get_option('opcounter'):
            if type_context == "f":
                _type = _type_to_op_counter_type('float')
                ret = Literal("{}({})".format(_type, ret.s))
            if type_context == "d":
                _type = _type_to_op_counter_type('double')
                ret = Literal("{}({})".format(_type, ret.s))
        return ret

    def map_fused_multiply_add(self, expr, type_context):
        if self.codegen_state.vectorization_info:
            # If this is vectorized we call the VCL function mul_add
            return prim.Call(prim.Variable("mul_add"),
                             (self.rec(expr.mul_op1), self.rec(expr.mul_op2), self.rec(expr.add_op)))
        else:
            # Default implementation that discards the node in favor of the resp.
            # additions and multiplications.
            return self.rec(expr.mul_op1 * expr.mul_op2 + expr.add_op, type_context)


class DuneCExpressionToCodeMapper(CExpressionToCodeMapper):
    def map_remainder(self, expr, enclosing_prec):
        n = expr.denominator

        # check whether n is a power of 2
        if isinstance(n, int) and n & n - 1 == 0:
            # Issue optimized code using bit masks
            from pymbolic.mapper.stringifier import PREC_PRODUCT
            x = self.rec(expr.numerator, PREC_PRODUCT)
            return "{} & {}".format(x, n - 1)
        else:
            return CExpressionToCodeMapper.map_remainder(expr, enclosing_prec)

    def map_min(self, expr, enclosing_prec):
        """ Max/Min is not implemented as a function!
        TODO: Revisit this w.r.t. ADL problems
        """
        what = type(expr).__name__.lower()

        children = list(set(expr.children))

        result = self.rec(children.pop(), PREC_NONE)
        while children:
            result = "std::%s(%s, %s)" % (what,
                                          self.rec(children.pop(), PREC_NONE),
                                          result)

        return result

    map_max = map_min


class DuneASTBuilder(CASTBuilder):
    def function_manglers(self):
        return CASTBuilder.function_manglers(self) + retrieve_cache_functions("mangler")

    def get_expression_to_c_expression_mapper(self, codegen_state):
        return DuneExpressionToCExpressionMapper(codegen_state)

    def get_c_expression_to_code_mapper(self):
        return DuneCExpressionToCodeMapper()

    def get_temporary_decl(self, knl, schedule_index, temp_var, decl_info):
        # If this is not a DuneTemporaryVariable, it was introduced by loopy
        # and it should be totally under loopys control: Call the base class implementation!
        if not (isinstance(temp_var, DuneTemporaryVariable) and temp_var.custom_declaration):
            return CASTBuilder.get_temporary_decl(self, knl, schedule_index, temp_var, decl_info)

        if temp_var.decl_method:
            return cgen.Line(temp_var.decl_method(temp_var.name, temp_var.shape, temp_var.shape_impl))

    def add_vector_access(self, access_expr, index):
        # There is no generic way of implementing a vector access with VCL, as
        # it might be that the entire statement needs to be rewritten. Consider
        # the example of an assignment to a vector component. It is *not* of the
        # form 'x.0 = 2' but instead its 'x.insert(0, 2)'. It is currently not
        # clear to me how this can be done, so I avoid the situation entirely.
        raise NotImplementedError()

    def emit_barrier(self, kind, comment):
        post_include("#define BARRIER asm volatile(\"\": : :\"memory\")", filetag="operatorfile")
        return cgen.Line("BARRIER;")

    def get_temporary_decls(self, codegen_state, schedule_index):
        if self.target.declare_temporaries:
            return CASTBuilder.get_temporary_decls(self, codegen_state, schedule_index)
        else:
            return []


class DuneTarget(TargetBase):
    def __init__(self, declare_temporaries=True):
        # Set fortran_abi to allow reusing CASTBuilder for the moment
        self.fortran_abi = False
        self.declare_temporaries = declare_temporaries

    def split_kernel_at_global_barriers(self):
        return False

    def get_host_ast_builder(self):
        return DummyHostASTBuilder(self)

    def get_device_ast_builder(self):
        return DuneASTBuilder(self)

    def dtype_to_typename(self, dtype):
        if dtype.dtype.kind == "V":
            include_file("dune/perftool/vectorclass/vectorclass.h", filetag="operatorfile")
            return VCLTypeRegistry.names[dtype.dtype]
        else:
            return numpy_to_cpp_dtype(dtype.dtype.name)

    def is_vector_dtype(self, dtype):
        return False

    def vector_dtype(self, base, count):
        return NumpyType(VCLTypeRegistry.types[base.numpy_dtype, count],
                         target=self)
